/*********************************************\
*   BeApple2 : Apple ][+ emulator for BeOS    *
*                                             *
*        cpu.cpp : 6502 processor             *
\*********************************************/

#include "cpu.h"
#include "bus.h"
#include "disk.h"
#include "speaker.h"
#include "tape.h"
#include <OS.h>
#include <stdio.h>
#include <KernelKit.h>

extern bool Sniff;
FILE *Fil=NULL;

thread_id ThreadCPU=0;

// CONFIG VARIABLES ---------
int Speed=1000;		// Vitesse CPU (kHz)
int LoopCycles=2000;	// Instructions / bloc
bool CMOS=false;	// 6502=false 65C02=true

// CPU REGISTERS ------------
const uint8 _flagC=1,_flagZ=2,_flagI=4,_flagD=8,_flagB=16,_flagR=32,_flagV=64,_flagN=128;
vuint8  GrA,GrX,GrY;                   // Accu. Index X,Y
vuint8  GrSP;      	                  // Stack
vuint16 GrPC;        	              // Prog. Counter 
vuint8  GrP;             	          // Status register

// CPU task Control ---------
volatile bool StopReq=false;	// Start/stop CPU
volatile bool StopAck=false;	// Handshake
volatile bool StopCPU=false;	// Final END

bool Pause=false;
// Siulation timing ---------
int64 SimCycles=0;	// Cycles simulation
int64 TimCycles=0;	// Cycles temporisation
int	FasCycles=0;	// Cycles boucle interne
vint64 StartTime;	// Emulation start time

//---------------------------
bool UpIt=false;	// Mise à jour registres Généraux

#define Bopr(a) (MEM[a])
#define Bopw(a,b) { MEM[(uint16)a]=b; }

#define BUSR(a) ( ( (((a)&0xFF00)!=0xC000) )?MEM[(a)]:Busr(a))
#define BUSW Busw

//##############################################################################################

int InstrCycles[256]= {
//	0	1	2	3	4	5	6	7		8	9	A	B	C	D	E	F
	 7	,6	,99	,99	,99	,3	,5	,99		,3	,2	,2	,99	,99	,4	,6	,99	// 0
	,2	,5	,99	,99	,99	,4	,6	,99		,2	,4	,99	,99	,99	,4	,7	,99	// 1
	,6	,6	,99	,99	,3	,3	,5	,99		,4	,2	,2	,99	,4	,4	,6	,99	// 2
	,2	,5	,99	,99	,99	,4	,6	,99		,2	,4	,99	,99	,99	,4	,7	,99	// 3
	,6	,6	,99	,99	,99	,3	,5	,99		,3	,2	,2	,99	,3	,4	,6	,99	// 4
	,2	,5	,99	,99	,99	,4	,6	,99		,2	,4	,99	,99	,99	,4	,7	,99	// 5
	,6	,6	,99	,99	,99	,3	,5	,99		,4	,2	,2	,99	,5	,4	,6	,99	// 6
	,2	,5	,99	,99	,99	,4	,6	,99		,2	,4	,99	,99	,99	,4	,7	,99	// 7
	,99	,6	,99	,99	,3	,3	,3	,99		,2	,99	,2	,99	,4	,4	,4	,99	// 8
	,2	,5	,99	,99	,4	,4	,4	,99		,2	,4	,2	,99	,99	,4	,99	,99	// 9
	,2	,6	,2	,99	,3	,3	,3	,99		,2	,2	,2	,99	,4	,4	,4	,99	// A
	,2	,5	,99	,99	,4	,4	,4	,99		,2	,4	,2	,99	,4	,4	,4	,99	// B
	,2	,6	,99	,99	,3	,3	,5	,99		,2	,2	,2	,99	,4	,4	,6	,99	// C
	,2	,5	,99	,99	,99	,4	,6	,99		,2	,4	,99	,99	,99	,4	,7	,99	// D
	,2	,6	,99	,99	,3	,3	,5	,99		,2	,2	,2	,99	,4	,4	,6	,99	// E
	,2	,5	,99	,99	,99	,4	,6	,99		,2	,4	,99	,99	,99	,4	,7	,99	// F
	};

//##############################################################################################

/*
## NOTE : Not so good ideas :
## - Remplacing #define-s below by modern inline functions 
## - Using bit fields like : union {	uint8  rP;
##										struct Tflag { bool C:1; bool Z:1; bool I:1; ... } flag; };
## Resulting increased executable size and decreased speed.
##*/


/* MODES D'ADRESSAGE */


#define VAL8	{ W=MEM[rPC++]; }
#define VAL16	{ W=(uint16)MEM[rPC]|((uint16)((uint16)MEM[rPC+1]<<8)); rPC+=2; }

#define IMM		{ V=MEM[rPC++]; }

#define ZPGX	{ W=(uint8)(MEM[rPC++]+rX); }
#define ZPGY	{ W=(uint8)(MEM[rPC++]+rY); }

#define ABSX	{ W=((uint16)MEM[rPC]|((uint16)((uint16)MEM[rPC+1]<<8))); W+=(uint16)rX; rPC+=2; }
#define ABSY	{ W=((uint16)MEM[rPC]|((uint16)((uint16)MEM[rPC+1]<<8))); W+=(uint16)rY; rPC+=2; }

#define IND		{ V=MEM[rPC++]; W=(uint16)MEM[V]+(uint16)MEM[V+1]*256; }
#define INDX	{ V=MEM[rPC++]+rX; W=(uint16)MEM[V]+(uint16)MEM[V+1]*256; }
#define INDY	{ V=MEM[rPC++]; W=(uint16)MEM[V]+(uint16)MEM[V+1]*256; W+=rY; }

#define PUSH(a); { MEM[256+(rSP--)]=(uint8)(a); }
#define PULL(a); { rSP++; (a)=MEM[256+rSP]; }

#define SETZ(a);  { rP&=~_flagZ; if ((a)==0) rP|=_flagZ; }
#define SETZN(a); { rP&=~(_flagZ|_flagN); if (!(a)) rP|=_flagZ; if ((a)&0x80) rP|=_flagN; }

#define Signe(a) ((int16)((int8)a))

/* INSTRUCTIONS */

#define ORA { rA|=V; rP&=~(_flagZ|_flagN); if (rA==0) rP|=_flagZ; if (rA&0x80) rP|=_flagN;}
#define AND { rA&=V; rP&=~(_flagZ|_flagN); if (rA==0) rP|=_flagZ; if (rA&0x80) rP|=_flagN;}
#define EOR { rA^=V; rP&=~(_flagZ|_flagN); if (rA==0) rP|=_flagZ; if (rA&0x80) rP|=_flagN;}

#define ADC { uint16 T; if (rP&_flagD) {										\
    	T=(uint16)((rA>>4)*10+(rA&15))+(uint16)((V>>4)*10+(V&15))+((rP&_flagC)?1:0);		\
		rP&=~(_flagC|_flagV);													\
		if (T>99) rP|=_flagC;													\
		if  (((rA&128)==(V&128))&&((rA&128)!=(T&128))) rP|=_flagV;				\
		rA=(((T/10)%10)<<4)+(T%10);												\
		/*printf ("ADC décimal !\n");*/												\
	  }else{																	\
		T=(uint16)rA+(uint16)V+((rP&_flagC)?1:0);								\
		rP&=~(_flagC|_flagV);													\
		if (T>255) rP|=_flagC;													\
		if (((rA&128)==(V&128))&&((rA&128)!=(T&128))) rP|=_flagV;				\
		rA=T;																	\
 		}																		\
	rP&=~(_flagZ|_flagN); if (rA==0) rP|=_flagZ; if (rA&0x80) rP|=_flagN;		\
	}

#define SBC { uint16 T; if (rP&_flagD) { 										\
		/*printf ("SBC décimal ! rA=%X V=%X C=%i\n",(int)rA,(int)V,(int)(rP&_flagC));*/		\
		T=(uint16)((rA>>4)*10+(rA&15))-(uint16)((V>>4)*10+(V&15))-((rP&_flagC)?0:1);	\
		rP&=~(_flagC|_flagV);													\
		if  (((rA&128)!=(V&128))&&((rA&128)!=(T&128))) rP|=_flagV;				\
		if (T<100) rP|=_flagC; else T=100+T;									\
		rA=(((T/10)%10)<<4)+(T%10);												\
		/*printf ("SBC décimal ! %X C=%i\n",(int)rA,(int)(rP&_flagC));			*/\
	 }else{																		\
		T=(uint16)rA-(uint16)V-((rP&_flagC)?0:1);								\
		rP&=~(_flagC|_flagV);													\
		if (T<0x8000) rP|=_flagC;												\
		if  (((rA&128)!=(V&128))&&((rA&128)!=(T&128))) rP|=_flagV;				\
		rA=T;																	\
 		}																		\
	rP&=~(_flagZ|_flagN); if (rA==0) rP|=_flagZ; if (rA&0x80) rP|=_flagN;		\
	}

#define CMP(a) { rP&=~(_flagZ|_flagN|_flagC);						\
	if ((a)>=V) rP|=_flagC;											\
	if ((a)==V) rP|=_flagZ; if (((a)-V)&0x80) rP|=_flagN;			\
	}

#define ASL(a) { rP&=~(_flagZ|_flagN|_flagC);						\
	if (((a)&128)!=0) rP|=_flagC;									\
	(a)<<=1; if ((a)==0) rP|=_flagZ; if ((a)&0x80) rP|=_flagN;		\
	}

#define ROL(a) { bool flag=((a)&128); (a)<<=1;						\
	if (rP&_flagC) (a)++;	rP&=~(_flagZ|_flagN|_flagC); 			\
	if (flag) rP|=_flagC;											\
	if (!(a)) rP|=_flagZ; if ((a)&0x80) rP|=_flagN;					\
	}
	
#define ROR(a) { bool flag=(a)&1; (a)>>=1;							\
	if (rP&_flagC) (a)|=128; rP&=~(_flagZ|_flagN|_flagC); 			\
	if (flag) rP|=_flagC;											\
	if (!(a)) rP|=_flagZ; if ((a)&0x80) rP|=_flagN;					\
	}
	
#define LSR(a) {	rP&=~(_flagZ|_flagN|_flagC);					\
	if ((a)&1) rP|=_flagC; (a)>>=1;									\
	if (!(a)) rP|=_flagZ; if ((a)&0x80) rP|=_flagN;					\
	}

// Spécial CMOS
#define TRB {	rP&=~(_flagZ);		\
	if (!(V&rA)) rP|=_flagZ;		\
	V&=~rA;							\
	}

#define TSB {	rP&=~(_flagZ);		\
	if (!(V&rA)) rP|=_flagZ;		\
	V|=rA;							\
	}


// Invalid opcodes
#define CHAOS_1 { Invalid(IC); }
#define CHAOS_2 { rPC++; Invalid(IC); }
#define CHAOS_3 { rPC+=2; Invalid(IC); }

//##############################################################################################
void Invalid(int IC) { printf ("---!!! Instruction invalide !!!-%2X --\n",IC); }
 
void CPU(void)
{
 
// int I;
 uint8 V;
 uint16 W;
 uint8 IC;

 uint8  rA,rX,rY;            // Accu. Index X,Y
 uint8  rSP;      	                  // Stack
 uint16 rPC;                 // Prog. Counter 
 uint8  rP;                  // Status register

 rA=GrA; rX=GrX; rY=GrY;
 rPC=GrPC; rP=GrP; rSP=GrSP;
 FasCycles=0; 
 do {
	IC=Bopr(rPC++);

	switch (IC) {
	case 0x00 : rPC++; PUSH(rPC>>8); PUSH(rPC); rP|=_flagB; PUSH(rP|_flagR);
			rP|=_flagI; W=0xFFFE; rPC=BUSR(W); rPC|=(uint16)BUSR(W+1)*256;	FasCycles+=InstrCycles[0x00]; break;	// BRK
    case 0x01 : INDX; V=BUSR(W); ORA; 										FasCycles+=InstrCycles[0x01]; break;	// ORA indX 
	case 0x02 : CHAOS_2;                                                    FasCycles+=InstrCycles[0x02]; break;
	case 0x03 : CHAOS_2;                                                    FasCycles+=InstrCycles[0x03]; break;
	case 0x04 : if (CMOS) {VAL8; V=Bopr(W); TSB; Bopw(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0x04]; break;  // TSB zpg (CMOS)
	case 0x05 : VAL8; V=Bopr(W); ORA;										FasCycles+=InstrCycles[0x05]; break;	// ORA zpg
	case 0x06 : VAL8; V=Bopr(W); ASL(V); Bopw(W,V);							FasCycles+=InstrCycles[0x06]; break;	// ASL zpg
	case 0x07 : CHAOS_2;                                                    FasCycles+=InstrCycles[0x07]; break;
	case 0x08 : PUSH(rP|_flagR|_flagB); /*Flag B present ?*/				FasCycles+=InstrCycles[0x08]; break;	// PHP imp 
	case 0x09 : IMM; ORA;											 		FasCycles+=InstrCycles[0x09]; break;	// ORA imm
	case 0x0A : ASL(rA);												 	FasCycles+=InstrCycles[0x0A]; break;	// ASL Acc
	case 0x0B : CHAOS_2;                                                    FasCycles+=InstrCycles[0x0B]; break;
	case 0x0C : if (CMOS) { VAL16; V=BUSR(W); TSB; BUSW(W,V); } else CHAOS_3;FasCycles+=InstrCycles[0x0C]; break;	// TSB abs (CMOS)
	case 0x0D : VAL16; V=BUSR(W); ORA;										FasCycles+=InstrCycles[0x0D]; break;	// ORA abs
	case 0x0E : VAL16; V=BUSR(W); ASL(V); BUSW(W,V);						FasCycles+=InstrCycles[0x0E]; break;	// ASL abs
	case 0x0F : CHAOS_3;                                                    FasCycles+=InstrCycles[0x0F]; break;

	case 0x10 : IMM; if (!(rP&_flagN)) rPC+=Signe(V);						FasCycles+=InstrCycles[0x10]; break;	// BPL rel
	case 0x11 : INDY; V=BUSR(W); ORA;										FasCycles+=InstrCycles[0x11]; break;	// ORA indY
	case 0x12 : if (CMOS) { IND; V=BUSR(W); ORA; BUSW(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0x12]; break;	// ORA ind (CMOS)
	case 0x13 : CHAOS_2;                                                    FasCycles+=InstrCycles[0x13]; break;
	case 0x14 : if (CMOS) {VAL8; V=Bopr(W); TRB; Bopw(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0x14]; break;  // TRB zpg (CMOS)
	case 0x15 : ZPGX; V=Bopr(W); ORA;										FasCycles+=InstrCycles[0x15]; break;	// ORA zpgX
	case 0x16 : ZPGX; V=Bopr(W); ASL(V); Bopw(W,V);							FasCycles+=InstrCycles[0x16]; break;	// ASL zpgX
	case 0x17 : CHAOS_2;                                                    FasCycles+=InstrCycles[0x17]; break;
	case 0x18 : rP&=~_flagC;												FasCycles+=InstrCycles[0x18]; break;	// CLC imp
	case 0x19 : ABSY; V=BUSR(W); ORA;										FasCycles+=InstrCycles[0x19]; break;	// ORA absY
	case 0x1A : if (CMOS) { rA++; SETZN(rA); } else CHAOS_1;				FasCycles+=InstrCycles[0x1A]; break;  // INC Acc (CMOS)
	case 0x1B : CHAOS_2;                                                    FasCycles+=InstrCycles[0x1B]; break;
	case 0x1C : if (CMOS) { VAL16; V=BUSR(W); TRB; BUSW(W,V); } else CHAOS_3;FasCycles+=InstrCycles[0x1C]; break;	// TRB abs (CMOS)
	case 0x1D : ABSX; V=BUSR(W); ORA;										FasCycles+=InstrCycles[0x1D]; break;	// ORA absX
	case 0x1E : ABSX; V=BUSR(W); ASL(V); BUSW(W,V);							FasCycles+=InstrCycles[0x1E]; break;	// ASL absX
	case 0x1F : CHAOS_3;													FasCycles+=InstrCycles[0x1F]; break;

	case 0x20 : VAL16; rPC--; PUSH(rPC>>8); PUSH(rPC); rPC=W;				FasCycles+=InstrCycles[0x20]; break;	// JSR
	case 0x21 : INDX; V=BUSR(W); AND;										FasCycles+=InstrCycles[0x21]; break;	// AND indX
	case 0x22 : CHAOS_2;                            						FasCycles+=InstrCycles[0x22]; break;
	case 0x23 : CHAOS_2;													FasCycles+=InstrCycles[0x23]; break;
	case 0x24 : VAL8; V=Bopr(W); rP&=~(_flagZ|_flagN|_flagV); 
		if (!(rA&V)) rP|=_flagZ; if (V&0x80) rP|=_flagN;
		if (V&0x40) rP|=_flagV;												FasCycles+=InstrCycles[0x24]; break;	// BIT zpg
	case 0x25 : VAL8; V=Bopr(W); AND;										FasCycles+=InstrCycles[0x25]; break;	// AND zpg
	case 0x26 : VAL8; V=Bopr(W); ROL(V); Bopw(W,V);							FasCycles+=InstrCycles[0x26]; break;	// ROL zpg
	case 0x27 : CHAOS_2;													FasCycles+=InstrCycles[0x27]; break;
	case 0x28 : PULL(rP);													FasCycles+=InstrCycles[0x28]; break;	// PLP imp
	case 0x29 : IMM; AND;													FasCycles+=InstrCycles[0x29]; break;	// AND imm
	case 0x2A : ROL(rA);													FasCycles+=InstrCycles[0x2A]; break;	// ROL Acc
	case 0x2B : CHAOS_2;													FasCycles+=InstrCycles[0x2B]; break;
	case 0x2C : VAL16; V=BUSR(W); rP&=~(_flagZ|_flagN|_flagV); 
		if (!(rA&V)) rP|=_flagZ; if (V&0x80) rP|=_flagN;
		if (V&0x40) rP|=_flagV;												FasCycles+=InstrCycles[0x2C]; break;	// BIT abs
	case 0x2D : VAL16; V=BUSR(W); AND;										FasCycles+=InstrCycles[0x2D]; break;	// AND abs
	case 0x2E : VAL16; V=BUSR(W); ROL(V); BUSW(W,V);						FasCycles+=InstrCycles[0x2E]; break;	// ROL abs
	case 0x2F : CHAOS_3;													FasCycles+=InstrCycles[0x2F]; break;

	case 0x30 : IMM; if (rP&_flagN) rPC+=Signe(V);							FasCycles+=InstrCycles[0x30]; break;	// BMI rel
	case 0x31 : INDY; V=BUSR(W); AND;										FasCycles+=InstrCycles[0x31]; break;	// AND indY
	case 0x32 : if (CMOS) { IND; V=BUSR(W); AND; BUSW(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0x32]; break;	// AND ind (CMOS)
	case 0x33 : CHAOS_2;													FasCycles+=InstrCycles[0x33]; break;
	case 0x34 : if (CMOS) { ZPGX; V=Bopr(W); rP&=~(_flagZ|_flagN|_flagV); 
		if (!(rA&V)) rP|=_flagZ; if (V&0x80) rP|=_flagN;
		if (V&0x40) rP|=_flagV; } else CHAOS_2;								FasCycles+=InstrCycles[0x34]; break;	// BIT zpgX
	case 0x35 : ZPGX; V=Bopr(W); AND;										FasCycles+=InstrCycles[0x35]; break;	// AND zpgX
	case 0x36 : ZPGX; V=Bopr(W); ROL(V); Bopw(W,V);							FasCycles+=InstrCycles[0x36]; break;	// ROL zpgX
	case 0x37 : CHAOS_2;													FasCycles+=InstrCycles[0x37]; break;
	case 0x38 : rP|=_flagC;													FasCycles+=InstrCycles[0x38]; break;	// SEC imp
	case 0x39 : ABSY; V=BUSR(W); AND;										FasCycles+=InstrCycles[0x39]; break;	// AND absY
	case 0x3A : if (CMOS) { rA--; SETZN(rA); } else CHAOS_1;				FasCycles+=InstrCycles[0x3A]; break;  // DEC Acc (CMOS)
	case 0x3B : CHAOS_2;													FasCycles+=InstrCycles[0x3B]; break;
	case 0x3C :if (CMOS) { ABSX; V=BUSR(W); rP&=~(_flagZ|_flagN|_flagV); 
		if (!(rA&V)) rP|=_flagZ; if (V&0x80) rP|=_flagN;
		if (V&0x40) rP|=_flagV; } else CHAOS_3;								FasCycles+=InstrCycles[0x3C]; break;	// BIT absX
	case 0x3D : ABSX; V=BUSR(W); AND;										FasCycles+=InstrCycles[0x3D]; break;	// AND absX
	case 0x3E : ABSX; V=BUSR(W); ROL(V); BUSW(W,V);							FasCycles+=InstrCycles[0x3E]; break;	// ROL absX
	case 0x3F : CHAOS_3;													FasCycles+=InstrCycles[0x3F]; break;

	case 0x40 : PULL(rP); PULL(rPC); PULL(V); rPC|=(uint16)V*256;			FasCycles+=InstrCycles[0x40]; break;	// RTI imp
	case 0x41 : INDX; V=BUSR(W); EOR;										FasCycles+=InstrCycles[0x41]; break;	// XOR indX
	case 0x42 : CHAOS_2;													FasCycles+=InstrCycles[0x42]; break;
	case 0x43 : CHAOS_2;													FasCycles+=InstrCycles[0x43]; break;
	case 0x44 : CHAOS_2;													FasCycles+=InstrCycles[0x44]; break;
	case 0x45 : VAL8; V=Bopr(W); EOR;										FasCycles+=InstrCycles[0x45]; break;	// EOR zpg
	case 0x46 : VAL8; V=Bopr(W); LSR(V); Bopw(W,V);							FasCycles+=InstrCycles[0x46]; break;	// LSR zpg
	case 0x47 : CHAOS_2;													FasCycles+=InstrCycles[0x47]; break;
	case 0x48 : PUSH(rA);													FasCycles+=InstrCycles[0x48]; break;	// PHA imp
	case 0x49 : IMM; EOR;													FasCycles+=InstrCycles[0x49]; break;	// EOR imm
	case 0x4A : LSR(rA);													FasCycles+=InstrCycles[0x4A]; break;	// LSR acc
	case 0x4B : CHAOS_2;													FasCycles+=InstrCycles[0x4B]; break;
	case 0x4C : VAL16; rPC=W;												FasCycles+=InstrCycles[0x4C]; break;	// JMP abs
	case 0x4D : VAL16; V=BUSR(W); EOR;										FasCycles+=InstrCycles[0x4D]; break;	// EOR abs
	case 0x4E : VAL16; V=BUSR(W); LSR(V); BUSW(W,V);						FasCycles+=InstrCycles[0x4E]; break;	// LSR abs
	case 0x4F : CHAOS_3;													FasCycles+=InstrCycles[0x4F]; break;

	case 0x50 : IMM; if (!(rP&_flagV)) rPC+=Signe(V);						FasCycles+=InstrCycles[0x50]; break;	// BVC rel
	case 0x51 : INDY; V=BUSR(W); EOR;										FasCycles+=InstrCycles[0x51]; break;	// EOR indY
	case 0x52 : if (CMOS) { IND; V=BUSR(W); EOR; BUSW(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0x52]; break;	// EOR ind (CMOS)
	case 0x53 : CHAOS_2;													FasCycles+=InstrCycles[0x53]; break;
	case 0x54 : CHAOS_2;													FasCycles+=InstrCycles[0x54]; break;
	case 0x55 : ZPGX; V=Bopr(W); EOR;										FasCycles+=InstrCycles[0x55]; break;	// EOR zpgX 
	case 0x56 : ZPGX; V=Bopr(W); LSR(V); Bopw(W,V);							FasCycles+=InstrCycles[0x56]; break;	// LSR zpgX
	case 0x57 : CHAOS_2;													FasCycles+=InstrCycles[0x57]; break;
	case 0x58 : rP&=~_flagI;												FasCycles+=InstrCycles[0x58]; break;	// CLI imp
	case 0x59 : ABSY; V=BUSR(W); EOR;										FasCycles+=InstrCycles[0x59]; break;	// EOR absY
	case 0x5A : if (CMOS) PUSH(rY);/* else CHAOS_1;		*/					FasCycles+=InstrCycles[0x5A]; break;	// PHY imp
	case 0x5B : CHAOS_2;													FasCycles+=InstrCycles[0x5B]; break;
	case 0x5C : CHAOS_3;													FasCycles+=InstrCycles[0x5C]; break;
	case 0x5D : ABSX; V=BUSR(W); EOR;										FasCycles+=InstrCycles[0x5D]; break;	// EOR absX
	case 0x5E : ABSX; V=BUSR(W); LSR(V); BUSW(W,V);							FasCycles+=InstrCycles[0x5E]; break;	// LSR absX
	case 0x5F : CHAOS_3;													FasCycles+=InstrCycles[0x5F]; break;

	case 0x60 : PULL(rPC); PULL(V); rPC|=(uint16)V*256; rPC++;				FasCycles+=InstrCycles[0x60]; break;	// RTS imp
	case 0x61 : INDX; V=BUSR(W); ADC;										FasCycles+=InstrCycles[0x61]; break;	// ADC indX
	case 0x62 : CHAOS_2;													FasCycles+=InstrCycles[0x62]; break;
	case 0x63 : CHAOS_2;													FasCycles+=InstrCycles[0x63]; break;
	case 0x64 : if (CMOS) { VAL8; Bopw(W,0); } else CHAOS_2;				FasCycles+=InstrCycles[0x64]; break;	// STZ zpg
	case 0x65 : VAL8; V=Bopr(W); ADC;										FasCycles+=InstrCycles[0x65]; break;	// ADC zpg
	case 0x66 : VAL8; V=Bopr(W); ROR(V); Bopw(W,V);							FasCycles+=InstrCycles[0x66]; break;	// ROR zpg
	case 0x67 : CHAOS_2;													FasCycles+=InstrCycles[0x67]; break;
	case 0x68 : PULL(rA); SETZN(rA);										FasCycles+=InstrCycles[0x68]; break;	// PLA imp
	case 0x69 : IMM; ADC;													FasCycles+=InstrCycles[0x69]; break;	// ADC imm
	case 0x6A : ROR(rA);													FasCycles+=InstrCycles[0x6A]; break;	// ROR acc
	case 0x6B : CHAOS_2;													FasCycles+=InstrCycles[0x6B]; break;
	case 0x6C : VAL16; rPC=BUSR(W);
		if ((!CMOS)&&((W&255)==255)) { printf ("Erreur JMP indirect !!!!\n"); 
							rPC|=(uint16)BUSR(W+1-256)*256; } else
		rPC|=(uint16)BUSR(W+1)*256;											FasCycles+=InstrCycles[0x6C]; break;	// JMP ind
	case 0x6D : VAL16; V=BUSR(W); ADC;										FasCycles+=InstrCycles[0x6D]; break;	// ADC abs
	case 0x6E : VAL16; V=BUSR(W); ROR(V); BUSW(W,V);						FasCycles+=InstrCycles[0x6E]; break;	// ROR abs
	case 0x6F : CHAOS_3;													FasCycles+=InstrCycles[0x6F]; break;

	case 0x70 : IMM; if (rP&_flagV) rPC+=Signe(V);							FasCycles+=InstrCycles[0x70]; break;	// BVS rel
	case 0x71 : INDY; V=BUSR(W); ADC;										FasCycles+=InstrCycles[0x71]; break;	// ADC indY
	case 0x72 : if (CMOS) { IND; V=BUSR(W); ADC; BUSW(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0x72]; break;	// ADC ind (CMOS)
	case 0x73 : CHAOS_2;													FasCycles+=InstrCycles[0x73]; break;
	case 0x74 : if (CMOS) { ZPGX; Bopw(W,0); } else CHAOS_2;				FasCycles+=InstrCycles[0x74]; break;	// STZ zpgX (CMOS)
	case 0x75 : ZPGX; V=Bopr(W); ADC;										FasCycles+=InstrCycles[0x75]; break;	// ADC zpgX
	case 0x76 : ZPGX; V=Bopr(W); ROR(V); Bopw(W,V);							FasCycles+=InstrCycles[0x76]; break;	// ROR zpgX
	case 0x77 : CHAOS_2;													FasCycles+=InstrCycles[0x77]; break;
	case 0x78 : rP|=_flagI;													FasCycles+=InstrCycles[0x78]; break;	// SEI imp
	case 0x79 : ABSY; V=BUSR(W); ADC;										FasCycles+=InstrCycles[0x79]; break;	// ADC absY
	case 0x7A : if (CMOS) { PULL(rY); SETZN(rY); } else CHAOS_1;			FasCycles+=InstrCycles[0x7A]; break;	// PLY imp (CMOS)
	case 0x7B : CHAOS_2;													FasCycles+=InstrCycles[0x7B]; break;
	case 0x7C : if (CMOS) { VAL16; rPC=BUSR(W);
		rPC|=(uint16)BUSR(W+1)*256; rPC+=rX;	} else CHAOS_3;				FasCycles+=InstrCycles[0x7C]; break;	// JMP indX (CMOS)
	case 0x7D : ABSX; V=BUSR(W); ADC;										FasCycles+=InstrCycles[0x7D]; break;	// ADC absX
	case 0x7E : ABSX; V=BUSR(W); ROR(V); BUSW(W,V);							FasCycles+=InstrCycles[0x7E]; break;	// ROR absX
	case 0x7F : CHAOS_3;													FasCycles+=InstrCycles[0x7F]; break;

	case 0x80 : if (CMOS) { IMM; rPC+=Signe(V); } else CHAOS_2;				FasCycles+=InstrCycles[0x80]; break;	// BRA rel (CMOS)
	case 0x81 : INDX; BUSW(W,rA);											FasCycles+=InstrCycles[0x81]; break;	// STA indX
	case 0x82 : CHAOS_2;													FasCycles+=InstrCycles[0x82]; break;
	case 0x83 : CHAOS_2;													FasCycles+=InstrCycles[0x83]; break;
	case 0x84 : VAL8; Bopw(W,rY);											FasCycles+=InstrCycles[0x84]; break;	// STY zpg
	case 0x85 : VAL8; Bopw(W,rA);											FasCycles+=InstrCycles[0x85]; break;	// STA zpg
	case 0x86 : VAL8; Bopw(W,rX);											FasCycles+=InstrCycles[0x86]; break;	// STX zpg
	case 0x87 : CHAOS_2;													FasCycles+=InstrCycles[0x87]; break;
	case 0x88 : rY--; SETZN(rY);											FasCycles+=InstrCycles[0x88]; break;	// DEY imp
	case 0x89 : if (CMOS) {IMM; rP&=~(_flagZ|_flagN|_flagV); 
		if (!(rA&V)) rP|=_flagZ; if (V&0x80) rP|=_flagN;
		if (V&0x40) rP|=_flagV; } else CHAOS_2;								FasCycles+=InstrCycles[0x89]; break;	// BIT imm
	case 0x8A : rA=rX; SETZN(rA);											FasCycles+=InstrCycles[0x8A]; break;	// TXA imp
	case 0x8B : CHAOS_2;													FasCycles+=InstrCycles[0x8B]; break;
	case 0x8C : VAL16; BUSW(W,rY);											FasCycles+=InstrCycles[0x8C]; break;	// STY abs
	case 0x8D : VAL16; BUSW(W,rA);											FasCycles+=InstrCycles[0x8D]; break;	// STA abs
	case 0x8E : VAL16; BUSW(W,rX);											FasCycles+=InstrCycles[0x8E]; break;	// STX abs
	case 0x8F : CHAOS_3;													FasCycles+=InstrCycles[0x8F]; break;
	
	case 0x90 : IMM; if (!(rP&_flagC)) rPC+=Signe(V);						FasCycles+=InstrCycles[0x90]; break;	// BCC rel
	case 0x91 : INDY; BUSW(W,rA);											FasCycles+=InstrCycles[0x91]; break;	// STA indY
	case 0x92 : if (CMOS) { IND; BUSW(W,rA); } else CHAOS_2;				FasCycles+=InstrCycles[0x92]; break;	// STA ind (CMOS)
	case 0x93 : CHAOS_2;													FasCycles+=InstrCycles[0x93]; break;
	case 0x94 : ZPGX; Bopw(W,rY);											FasCycles+=InstrCycles[0x94]; break;	// STY zpgX 
	case 0x95 : ZPGX; Bopw(W,rA);											FasCycles+=InstrCycles[0x95]; break;	// STA zpgX 
	case 0x96 : ZPGY; Bopw(W,rX);											FasCycles+=InstrCycles[0x96]; break;	// STX zpgY 
	case 0x97 : CHAOS_2;													FasCycles+=InstrCycles[0x97]; break;
	case 0x98 : rA=rY; SETZN(rA);											FasCycles+=InstrCycles[0x98]; break;	// TYA imp
	case 0x99 : ABSY; BUSW(W,rA);											FasCycles+=InstrCycles[0x99]; break;	// STA absY
	case 0x9A : rSP=rX;														FasCycles+=InstrCycles[0x9A]; break;	// TXS imp
	case 0x9B : CHAOS_2;													FasCycles+=InstrCycles[0x9B]; break;
	case 0x9C : if (CMOS) { VAL16; BUSW(W,0); } else CHAOS_3;				FasCycles+=InstrCycles[0x9C]; break;	// STZ abs (CMOS)
	case 0x9D : ABSX; BUSW(W,rA);											FasCycles+=InstrCycles[0x9D]; break;	// STA absX
	case 0x9E : if (CMOS) { ABSX; BUSW(W,0); } else CHAOS_3;				FasCycles+=InstrCycles[0x9E]; break;	// STZ absX (CMOS)
	case 0x9F : CHAOS_3;													FasCycles+=InstrCycles[0x9F]; break;
	
	case 0xA0 : IMM; rY=V; SETZN(rY);										FasCycles+=InstrCycles[0xA0]; break;	// LDY imm
	case 0xA1 : INDX; rA=BUSR(W); SETZN(rA);								FasCycles+=InstrCycles[0xA1]; break;	// LDA indX 
	case 0xA2 : IMM; rX=V; SETZN(rX);										FasCycles+=InstrCycles[0xA2]; break;	// LDX imm
	case 0xA3 : CHAOS_2;													FasCycles+=InstrCycles[0xA3]; break;
	case 0xA4 : VAL8; rY=Bopr(W); SETZN(rY);								FasCycles+=InstrCycles[0xA4]; break;	// LDY zpg
	case 0xA5 : VAL8; rA=Bopr(W); SETZN(rA);								FasCycles+=InstrCycles[0xA5]; break;	// LDA zpg
	case 0xA6 : VAL8; rX=Bopr(W); SETZN(rX);								FasCycles+=InstrCycles[0xA6]; break;	// LDX zpg
	case 0xA7 : CHAOS_2;													FasCycles+=InstrCycles[0xA7]; break;
	case 0xA8 : rY=rA; SETZN(rY);											FasCycles+=InstrCycles[0xA8]; break;	// TAY imp
	case 0xA9 : IMM; rA=V; SETZN(rA);										FasCycles+=InstrCycles[0xA9]; break;	// LDA imm
	case 0xAA : rX=rA; SETZN(rX);											FasCycles+=InstrCycles[0xAA]; break;	// TAX imp
	case 0xAB : CHAOS_2;													FasCycles+=InstrCycles[0xAB]; break;
	case 0xAC : VAL16; rY=BUSR(W); SETZN(rY);								FasCycles+=InstrCycles[0xAC]; break;	// LDY abs
	case 0xAD : VAL16; rA=BUSR(W); SETZN(rA);								FasCycles+=InstrCycles[0xAD]; break;	// LDA abs
	case 0xAE : VAL16; rX=BUSR(W); SETZN(rX);								FasCycles+=InstrCycles[0xAE]; break;	// LDX abs
	case 0xAF : CHAOS_3;													FasCycles+=InstrCycles[0xAF]; break;
	  
	case 0xB0 : IMM; if (rP&_flagC) rPC+=Signe(V);							FasCycles+=InstrCycles[0xB0]; break;	// BCS rel
	case 0xB1 : INDY; rA=BUSR(W); SETZN(rA);								FasCycles+=InstrCycles[0xB1]; break;	// LDA indY
	case 0xB2 : if (CMOS) { IND; rA=BUSR(W); SETZN(rA); } else CHAOS_2;		FasCycles+=InstrCycles[0xB2]; break;	// LDA ind (CMOS)
	case 0xB3 : CHAOS_2;													FasCycles+=InstrCycles[0xB3]; break;
	case 0xB4 : ZPGX; rY=Bopr(W); SETZN(rY);								FasCycles+=InstrCycles[0xB4]; break;	// LDY zpgX 
	case 0xB5 : ZPGX; rA=Bopr(W); SETZN(rA);								FasCycles+=InstrCycles[0xB5]; break;	// LDA zpgX 
	case 0xB6 : ZPGY; rX=Bopr(W); SETZN(rX);								FasCycles+=InstrCycles[0xB6]; break;	// LDX zpgY 
	case 0xB7 : CHAOS_2;													FasCycles+=InstrCycles[0xB7]; break;
	case 0xB8 : rP&=~_flagV;												FasCycles+=InstrCycles[0xB8]; break;	// CLV imp
	case 0xB9 : ABSY; rA=BUSR(W); SETZN(rA);								FasCycles+=InstrCycles[0xB9]; break;	// LDA absY
	case 0xBA : rX=rSP; SETZN(rX);											FasCycles+=InstrCycles[0xBA]; break;	// TSX imp
	case 0xBB : CHAOS_2;													FasCycles+=InstrCycles[0xBB]; break;
	case 0xBC : ABSX; rY=BUSR(W); SETZN(rY);								FasCycles+=InstrCycles[0xBC]; break;	// LDY absX
	case 0xBD : ABSX; rA=BUSR(W); SETZN(rA);								FasCycles+=InstrCycles[0xBD]; break;	// LDA absX
	case 0xBE : ABSY; rX=BUSR(W); SETZN(rX);								FasCycles+=InstrCycles[0xBE]; break;	// LDX absY
	case 0xBF : CHAOS_3;													FasCycles+=InstrCycles[0xBF]; break;
	
	case 0xC0 : IMM; CMP(rY);												FasCycles+=InstrCycles[0xC0]; break;	// CPY imm
	case 0xC1 : INDX; V=BUSR(W); CMP(rA);									FasCycles+=InstrCycles[0xC1]; break;	// CMP indX
	case 0xC2 : CHAOS_2;													FasCycles+=InstrCycles[0xC2]; break;
	case 0xC3 : CHAOS_2;													FasCycles+=InstrCycles[0xC3]; break;
	case 0xC4 : VAL8; V=Bopr(W); CMP(rY);									FasCycles+=InstrCycles[0xC4]; break;	// CPY zpg
	case 0xC5 : VAL8; V=Bopr(W); CMP(rA);									FasCycles+=InstrCycles[0xC5]; break;	// CMP zpg
	case 0xC6 : VAL8; V=Bopr(W)-1; Bopw(W,V); SETZN(V);						FasCycles+=InstrCycles[0xC6]; break;	// DEC zpg
	case 0xC7 : CHAOS_2;													FasCycles+=InstrCycles[0xC7]; break;
	case 0xC8 : rY++; SETZN(rY);											FasCycles+=InstrCycles[0xC8]; break;	// INY imp
	case 0xC9 : IMM; CMP(rA);												FasCycles+=InstrCycles[0xC9]; break;	// CMP imm
	case 0xCA : rX--; SETZN(rX);											FasCycles+=InstrCycles[0xCA]; break;	// DEX imp
	case 0xCB : CHAOS_2;													FasCycles+=InstrCycles[0xCB]; break;
	case 0xCC : VAL16; V=BUSR(W); CMP(rY);									FasCycles+=InstrCycles[0xCC]; break;	// CPY abs
	case 0xCD : VAL16; V=BUSR(W); CMP(rA);									FasCycles+=InstrCycles[0xCD]; break;	// CMP abs
	case 0xCE : VAL16; V=BUSR(W)-1; BUSW(W,V); SETZN(V);					FasCycles+=InstrCycles[0xCE]; break;	// DEC abs
	case 0xCF : CHAOS_3;													FasCycles+=InstrCycles[0xCF]; break;

	case 0xD0 : IMM; if (!(rP&_flagZ)) rPC+=Signe(V);						FasCycles+=InstrCycles[0xD0]; break;	// BNE rel
	case 0xD1 : INDY; V=BUSR(W); CMP(rA);									FasCycles+=InstrCycles[0xD1]; break;	// CMP indY 
	case 0xD2 : if (CMOS) {IND;V=BUSR(W); CMP(rA); BUSW(W,V); } else CHAOS_2; FasCycles+=InstrCycles[0xD2]; break;	// CMP ind (CMOS)
	case 0xD3 : CHAOS_2;													FasCycles+=InstrCycles[0xD3]; break;
	case 0xD4 : CHAOS_2;													FasCycles+=InstrCycles[0xD4]; break;
	case 0xD5 : ZPGX; V=Bopr(W); CMP(rA);									FasCycles+=InstrCycles[0xD5]; break;	// CMP zpgX
	case 0xD6 : ZPGX; V=Bopr(W)-1; Bopw(W,V); SETZN(V);						FasCycles+=InstrCycles[0xD6]; break;	// DEC zpgX
	case 0xD7 : CHAOS_2;													FasCycles+=InstrCycles[0xD7]; break;
	case 0xD8 : rP&=~_flagD;												FasCycles+=InstrCycles[0xD8]; break;	// CLD imp
	case 0xD9 : ABSY; V=BUSR(W); CMP(rA);									FasCycles+=InstrCycles[0xD9]; break;	// CMP absY
	case 0xDA : if (CMOS) PUSH(rX); /*else CHAOS_1;*/						FasCycles+=InstrCycles[0xDA]; break;	// PHX imp
	case 0xDB : CHAOS_2;													FasCycles+=InstrCycles[0xDB]; break;
	case 0xDC : CHAOS_3;													FasCycles+=InstrCycles[0xDC]; break;
	case 0xDD : ABSX; V=BUSR(W); CMP(rA);									FasCycles+=InstrCycles[0xDD]; break;	// CMP absX
	case 0xDE : ABSX; V=BUSR(W)-1; BUSW(W,V); SETZN(V);						FasCycles+=InstrCycles[0xDE]; break;	// DEC absX
	case 0xDF : CHAOS_3;													FasCycles+=InstrCycles[0xDF]; break;

	case 0xE0 : IMM; CMP(rX);												FasCycles+=InstrCycles[0xE0]; break;	// CPX imm
	case 0xE1 : INDX; V=BUSR(W); SBC;										FasCycles+=InstrCycles[0xE1]; break;	// SBC indX
	case 0xE2 : CHAOS_2;													FasCycles+=InstrCycles[0xE2]; break;
	case 0xE3 : CHAOS_2;													FasCycles+=InstrCycles[0xE3]; break;
	case 0xE4 : VAL8; V=Bopr(W); CMP(rX);									FasCycles+=InstrCycles[0xE4]; break;	// CPX zpg
	case 0xE5 : VAL8; V=Bopr(W); SBC;										FasCycles+=InstrCycles[0xE5]; break;	// SBC zpg
	case 0xE6 : VAL8; V=Bopr(W)+1; Bopw(W,V); SETZN(V); 					FasCycles+=InstrCycles[0xE6]; break;	// INC zpg
	case 0xE7 : CHAOS_2;													FasCycles+=InstrCycles[0xE7]; break;
	case 0xE8 : rX++; SETZN(rX);											FasCycles+=InstrCycles[0xE8]; break;	// INX imp
	case 0xE9 : IMM; SBC;													FasCycles+=InstrCycles[0xE9]; break;	// SBC imm
	case 0xEA : /* NOP: The only universal assembly language instruction */	FasCycles+=InstrCycles[0xEA]; break;	// NOP imp
	case 0xEB : CHAOS_2;													FasCycles+=InstrCycles[0xEB]; break;
	case 0xEC : VAL16; V=BUSR(W); CMP(rX);									FasCycles+=InstrCycles[0xEC]; break;	// CPX abs
	case 0xED : VAL16; V=BUSR(W); SBC;										FasCycles+=InstrCycles[0xED]; break;	// SBC abs
	case 0xEE : VAL16; V=BUSR(W)+1; BUSW(W,V);  SETZN(V);					FasCycles+=InstrCycles[0xEE]; break;	// INC abs
	case 0xEF : CHAOS_3;													FasCycles+=InstrCycles[0xEF]; break;

	case 0xF0 : IMM; if (rP&_flagZ) rPC+=Signe(V);							FasCycles+=InstrCycles[0xF0]; break;	// BEQ rel
	case 0xF1 : INDY; V=BUSR(W); SBC;										FasCycles+=InstrCycles[0xF1]; break;	// SBC indY
	case 0xF2 : if (CMOS) { IND; V=BUSR(W); SBC; BUSW(W,V); } else CHAOS_2;	FasCycles+=InstrCycles[0xF2]; break;	// SBC ind (CMOS)
	case 0xF3 : CHAOS_2;													FasCycles+=InstrCycles[0xF3]; break;
	case 0xF4 : CHAOS_2;													FasCycles+=InstrCycles[0xF4]; break;
	case 0xF5 : ZPGX; V=Bopr(W); SBC;										FasCycles+=InstrCycles[0xF5]; break;	// SBC zpgX
	case 0xF6 : ZPGX; V=Bopr(W)+1; Bopw(W,V);  SETZN(V);					FasCycles+=InstrCycles[0xF6]; break;	// INC zpgX
	case 0xF7 : CHAOS_2;													FasCycles+=InstrCycles[0xF7]; break;
	case 0xF8 : rP|=_flagD;													FasCycles+=InstrCycles[0xF8]; break;	// SED imp
	case 0xF9 : ABSY; V=BUSR(W); SBC;										FasCycles+=InstrCycles[0xF9]; break;	// SBC absY
	case 0xFA : if (CMOS) { PULL(rX); SETZN(rX); } else CHAOS_1;			FasCycles+=InstrCycles[0xFA]; break;	// PLX imp (CMOS)
	case 0xFB : CHAOS_2;													FasCycles+=InstrCycles[0xFB]; break;
	case 0xFC : CHAOS_3;													FasCycles+=InstrCycles[0xFC]; break;
	case 0xFD : ABSX; V=BUSR(W); SBC;										FasCycles+=InstrCycles[0xFD]; break;	// SBC absX
	case 0xFE : ABSX; V=BUSR(W)+1; BUSW(W,V);  SETZN(V);					FasCycles+=InstrCycles[0xFE]; break;	// INC absX
	case 0xFF : CHAOS_3;													FasCycles+=InstrCycles[0xFF]; break;

	} // switch
 	} while (FasCycles<LoopCycles);
 GrA=rA;
 GrX=rX;
 GrY=rY;
 GrPC=rPC;
 GrP=rP;
 GrSP=rSP;
 
 SimCycles+=FasCycles;
 TimCycles+=FasCycles; FasCycles=0;
 return;
}

//---------------------------------------------------------------
int32 CPUTask(void *) // Tâche d'exécution du processeur !!!!!
{
 int64 C1,C2;
 do {
 	int64 SpkCycles=TimCycles;
	CPU();
	SpeakLoop(TimCycles-SpkCycles);
	if (StopReq) {
		StopAck=true;
		do { snooze(1000); } while (StopReq);
		StopAck=false;
		}
	ReadTape();
	
	C1=real_time_clock_usecs()/1000LL-StartTime;
	C2=TimCycles/Speed;
	
	if (C1<(C2-500000)) { TimCycles=C1*Speed; }
	else {
		while (C2>C1) {
			snooze(20000);
			C1=real_time_clock_usecs()/1000LL-StartTime;
			C2=TimCycles/Speed;
			}
		}
	} while (!StopCPU);
 return 0;
}

//###############################################################

//###############################################################

//###############################################################

int SwitchCPU(bool State)
{
 if (Pause!=State) return 0; // Already running / stopped
 if (State) { // Start & restart
	if (ThreadCPU) {
		StopReq=false;
		StopAck=false;
		StartTime=real_time_clock_usecs()/1000LL-TimCycles/Speed;
		resume_thread(ThreadCPU);
		Pause=false;
		} else return 1;
  }else{
	if (ThreadCPU) {
		resume_thread(ThreadCPU);
		StopAck=false;
		StopReq=true;
		do { snooze(500); } while (!StopAck);
		suspend_thread(ThreadCPU);
		Pause=true;
		} else return 1;
	}
 return 0;
}

//---------------------------------------------------------------
void InterCPU(int Type)
{
 SwitchCPU(false); // Pause
 switch(Type) {
 	case INTER_RESET:
 		GrA=GrX=GrY=0;
 		GrP=0x20;
 		GrPC=BUSR(0xFFFC)+BUSR(0xFFFD)*256;
 		//printf (" Reset Vector=%X\n",(int)GrPC);
 		GrSP=0xFF; //==0x1FF
 		Drive1.MotorON=false;
 		Drive1.UpMotor=true;
 		Drive2.MotorON=false;
 		Drive2.UpMotor=true;
  		break;
	case INTER_IRQ:
		break;
	case INTER_NMI:
		break;
	}
 SwitchCPU(true); // Restart
}

//---------------------------------------------------------------
void InitCPU()
{
 SimCycles=0;
 TimCycles=0;
 StartTime=real_time_clock_usecs()/1000LL;
 InterCPU(INTER_RESET);
}



